"""Wrap extract function logic to files."""
from __future__ import annotations

import errno
import os
from dataclasses import dataclass
from functools import reduce
from itertools import chain, groupby
from pathlib import Path
from typing import Any, Iterable

import nbformat

from nbdefs2py import funcs

NB_SUFFIX = ".ipynb"
PY_SUFFIX = ".py"


@dataclass(frozen=True)
class Definition:
    """Function representation from source file."""

    name: str
    path: Path
    src: str


class NotFoundError(FileNotFoundError):
    """Rename error since we're referring to file or directory."""


class ExistsError(FileExistsError):
    """Rename error since we're referring to file or directory."""


class FileSuffixError(Exception):
    """Invalid file extension."""

    def __init__(self: FileSuffixError, file: Path, message: str | None = None) -> None:
        """Print message and file."""
        super().__init__(f"{message or self.__doc__} Got `{file.suffix}`.")


class InputsError(ValueError):
    """Invalid inputs."""

    def __init__(self: InputsError, message: str | None = None) -> None:
        """Print custom message."""
        super().__init__(message or self.__doc__)


class PathNameError(Exception):
    """Path name does not match expected."""

    def __init__(self: PathNameError, src: Path, dest: Path) -> None:
        """Include path type (file or directory) and suffix."""
        super().__init__(
            f"'{str(src)}' is a {'directory' if src.is_dir() else 'file'}"
            f" and '{str(dest)}' has '{dest.suffix}' suffix."
            " If this is the desired behavior, pass `check_pathnames=False`",
        )


def _all_eq(iterable: Iterable[Any]) -> bool:
    """
    Check that all elements in iterable are equal.

    From `itertools` recipes. Check
     `https://docs.python.org/3/library/itertools.html#itertools-recipes` for more
     information.
    """
    g = groupby(iterable)
    return bool(next(g, True)) and not next(g, False)  # noqa: FBT003


def _combine_funcs(
    src: list[Definition],
    dest: list[Definition],
    update: bool | None,
) -> Iterable[Definition]:
    """
    Combine source and destination functions according to `update` strategy.

    :param src: source functions
    :param dest: destination functions
    :param update: update destination functions, overwrite or upsert (`True`, `False`
     and `None`, respectively)
    :return: merged list of functions
    """

    def _first_match(el: Definition, funcs: list[Definition]) -> Definition | None:
        """Return first match, if exists."""
        return next((f for f in funcs if f.name == el.name), None)

    funcs_update = dest if update in (True, None) else []
    updated = filter(
        lambda f: f is not None,
        (_first_match(func, src) for func in funcs_update),
    )
    if update is None:
        return chain.from_iterable((dest, set(src) - {f for f in updated if f}))
    return dest


def extract(
    src: Path | str,
    ignore: str | None = None,
    include: list[str] | None = None,
    exclude: list[str] | None = None,
    **read_kwargs: Any,  # noqa: ANN401
) -> list[Definition]:
    """
    Extract functions from `src` and write to `dest`.

    :param src: source location (file or directory)
    :param ignore: glob expression of files to ignore, defaults to `None`
    :param include: functions to include (`None` for all), defaults to `None`
    :param exclude: functions to exclude (`None` for none), defaults to `None`
    :param read_kwargs: keyword arguments to pass to `nbformat.read`
    :raises NotFoundError: when source is not found
    :raises FileSuffixError: when source file is not a notebook (`.ipynb`)
    :raises ExistsError: when destination exists and `overwrite=False`
    """
    src = Path(src)

    if not src.exists():
        raise NotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), src)
    if src.is_file() and src.suffix not in (PY_SUFFIX, NB_SUFFIX):
        raise FileSuffixError(src)

    if all(arg is not None for arg in (include, exclude)):
        raise InputsError("Must specify at least one of `include` or `exclude`.")

    ignore_glob = src.glob(ignore or "**/!*")
    paths = list(
        filter(
            lambda p: p not in ignore_glob,
            chain.from_iterable(src.rglob(f"*{s}") for s in (NB_SUFFIX, PY_SUFFIX))
            if src.is_dir()
            else [src],
        ),
    )

    if not _all_eq(path.suffix for path in paths):
        raise InputsError(
            "Expected only one file type, got"
            f" {list({path.suffix for path in paths})}.",
        )
    _src = {
        path: nbformat.read(path, **{"as_version": 4, **read_kwargs})
        if path.suffix == NB_SUFFIX
        else path.read_text()
        for path in paths
    }
    funcs_src = [
        Definition(
            name=fname,
            path=path,
            src=fsrc,
        )
        for path, node in _src.items()
        for fname, fsrc in reduce(
            lambda d1, d2: {**d1, **d2},
            funcs.from_obj(node),
            {},
        ).items()
    ]

    if all(arg is None for arg in (include, exclude)):
        return funcs_src
    keep = include or ({f.name for f in funcs_src} - set(exclude))  # pyre-ignore[6]
    return [func for func in funcs_src if func.name in keep]


def export(
    source: Path | str,
    destination: Path | str,
    *,
    update: bool | None = None,
    exist_ok: bool = True,
    check_pathnames: bool = True,
    **extract_kwargs: Any,  # noqa: ANN401
) -> None:
    """
    Export functions from notebook(s) `src` to `dest`.

    :param src: source location (file or directory)
    :param dest: destination location (file or directory)
    :param update: `True` only updates existing functions in destination, `False`
     overwrites and None will upsert functions, defaults to `None`
    :param exist_ok: allow existing destination file
    :param check_pathnames: check that files end in `.py` and directories do not
    :param extract_kwargs: keyword arguments to pass to `extract`
    """
    source = Path(source)
    destination = Path(destination)

    if check_pathnames and source.is_file() ^ bool(destination.suffix):
        raise PathNameError(source, destination)

    if destination.exists() and not exist_ok:
        raise ExistsError(
            errno.EEXIST,
            "Destination already exists and `exist_ok=False`.",
            destination,
        )

    if source.is_file():
        destination.touch(exist_ok=True)
    else:
        destination.mkdir(parents=True, exist_ok=True)

    funcs_src = extract(src=source, **extract_kwargs)
    funcs_dst = extract(src=destination)
    funcs_all = _combine_funcs(funcs_src, funcs_dst, update=update)

    for _path, _funcs in groupby(
        sorted(funcs_all, key=lambda e: e.path),
        key=lambda e: e.path,
    ):
        target = (
            (destination / _path.relative_to(source)).with_suffix(PY_SUFFIX)
            if source in (*_path.parents, _path)  # pyre-ignore[60]
            else _path
        )
        target.touch(exist_ok=True)
        target.write_text("\n\n".join(sorted(f.src for f in _funcs)))



"""Extract functions from notebooks."""
from __future__ import annotations

import ast
from dataclasses import asdict, dataclass
from typing import TypeVar

from nbformat import NotebookNode

DefNode = TypeVar("DefNode", ast.ClassDef, ast.FunctionDef)


@dataclass
class SubStrBound:
    """Bounds of a string to be extracted."""

    start_ln: int
    end_ln: int
    start_col: int
    end_col: int


class FuncExtractorStr(ast.NodeVisitor):
    """AST node visitor to extract functions from arbitrary text."""

    def __init__(self: FuncExtractorStr, src: str) -> None:
        """Add empty bounds - find when visiting AST tree."""
        self.func_bounds: dict[str, SubStrBound] = {}
        self.src = src

    @staticmethod
    def _substr(
        s: str,
        *,
        start_ln: int,
        end_ln: int,
        start_col: int,
        end_col: int,
    ) -> str:
        """Get a substring from start and end lines and columns."""
        lines = s.splitlines()[start_ln - 1 : end_ln]
        lines[0] = lines[0][start_col:]
        lines[-1] = lines[-1][:end_col]
        return "\n".join(lines)

    def visit_def(
        self: FuncExtractorStr,
        node: DefNode,
    ) -> DefNode:
        """Node visitor for class and function definitions."""
        self.func_bounds[node.name] = SubStrBound(
            start_ln=node.lineno,
            end_ln=node.end_lineno or -1,
            start_col=node.col_offset,
            end_col=node.end_col_offset or -1,
        )
        return node

    def visit_FunctionDef(  # noqa: N802
        self: FuncExtractorStr,
        node: ast.FunctionDef,
    ) -> ast.FunctionDef:
        """Add function string bounds to `self.func_bounds`."""
        return self.visit_def(node=node)

    def visit_ClassDef(  # noqa: N802
        self: FuncExtractorStr, node: ast.ClassDef
    ) -> ast.ClassDef:
        """Add class string bounds to `self.func_bounds`."""
        return self.visit_def(node=node)

    def funcs(self: FuncExtractorStr) -> dict[str, str]:
        """Extract the function definitions from the rest of the string."""
        self.visit(ast.parse(self.src))
        return {
            func: self._substr(self.src, **asdict(bound))
            for func, bound in self.func_bounds.items()
        }


def from_nb(nb: NotebookNode) -> list[dict[str, str]]:
    """Extract functions from notebook code cells."""
    return [
        FuncExtractorStr(cell.source).funcs()
        for cell in nb.cells
        if cell.cell_type == "code"
    ]


def from_src(src: str) -> dict[str, str]:
    """Extract functions from source file."""
    return FuncExtractorStr(src).funcs()


def from_obj(obj: str | NotebookNode) -> list[dict[str, str]]:
    """Extract functions from object (string or notebook)."""
    return [from_src(obj)] if isinstance(obj, str) else from_nb(obj)



"""Test extracting functions."""
from pathlib import Path

import nbformat

from nbdefs2py.funcs import FuncExtractorStr, from_nb

TEST_NB = Path(__file__).parent / "files" / "test.ipynb"


def test_extract_str() -> None:
    """Extract functions from a source string."""
    nb = nbformat.read(TEST_NB, as_version=4)
    assert FuncExtractorStr(nb.cells[0].source).funcs() == {
        "another_func": "def another_func() -> None:\n  ...",
        "print_this": "def print_this(x: str) -> None:\n  print(x)",
    }


def test_from_nb() -> None:
    """Extract functions from a notebook file."""
    nb = nbformat.read(TEST_NB, as_version=4)
    assert from_nb(nb) == [
        {
            "another_func": "def another_func() -> None:\n  ...",
            "print_this": "def print_this(x: str) -> None:\n  print(x)",
        },
        {},
        {
            "funcs_for_days": (
                "def funcs_for_days(x: str, i: int) -> str:\n"
                '  """Docstring."""\n  return "x: {x}; i: {i}"'
            ),
        },
        {
            "SomeClass": (
                "class SomeClass:\n"
                "    def __init__():\n"
                "        self.somethin = None\n\n"
                "    def gimme_somethin(self):\n"
                "        return self.somethin"
            )
        },
    ]



"""Tests for main CLI components."""
import contextlib
from pathlib import Path

import pytest  # pyre-ignore[21]
from _pytest.capture import CaptureFixture  # pyre-ignore[21]

from nbdefs2py.__main__ import main
from tests.test_io import FUNCS, NBPATH


@pytest.mark.parametrize("args", ["-h", "--help"])
def test_help(capsys: CaptureFixture, args: str) -> None:  # pyre-ignore[11]
    """Ensure module definition is correct in `--help`."""
    with contextlib.suppress(SystemExit):
        main([args])
    output = capsys.readouterr().out
    assert "Extract definitions from notebooks." in output


def test_cli(tmp_path: Path) -> None:
    """Ensure module definition is correct in `--help`."""
    target = tmp_path / "target.py"
    with contextlib.suppress(SystemExit):
        main([str(NBPATH), str(target)])
    assert target.read_text("utf-8") == "\n\n".join(FUNCS.values())



"""Test extracting functions."""
from __future__ import annotations

from pathlib import Path

import pytest  # pyre-ignore[21]

from nbdefs2py.io import (
    Definition,
    ExistsError,
    FileSuffixError,
    InputsError,
    NotFoundError,
    PathNameError,
    _combine_funcs,
    export,
    extract,
)

NBPATH = Path(__file__).parent / "files" / "test.ipynb"
FUNCS = {
    "SomeClass": (
        "class SomeClass:\n"
        "    def __init__():\n"
        "        self.somethin = None\n\n"
        "    def gimme_somethin(self):\n"
        "        return self.somethin"
    ),
    "another_func": "def another_func() -> None:\n  ...",
    "funcs_for_days": (
        'def funcs_for_days(x: str, i: int) -> str:\n  """Docstring."""\n'
        '  return "x: {x}; i: {i}"'
    ),
    "print_this": "def print_this(x: str) -> None:\n  print(x)",
}


@pytest.mark.parametrize(
    ("source", "target", "written"),
    [
        (NBPATH, "tmpfile.py", Path("tmpfile.py")),
        (
            NBPATH.parent,
            "tmpdir",
            Path("tmpdir") / NBPATH.with_suffix(".py").name,
        ),
    ],
)
def test_export(source: Path, target: str, written: Path, tmp_path: Path) -> None:
    """Export functions from a source to target file."""
    export(source, tmp_path / target)
    assert (tmp_path / written).read_text("utf-8") == "\n\n".join(FUNCS.values())


@pytest.mark.parametrize(
    ("source", "target", "written"),
    [
        (NBPATH, "tmpfile.py", Path("tmpfile.py")),
        (
            NBPATH.parent,
            "tmpdir",
            Path("tmpdir") / NBPATH.with_suffix(".py").name,
        ),
    ],
)
def test_export_overwrite(
    source: Path, target: str, written: Path, tmp_path: Path
) -> None:
    """Check that `exist_ok` raise errors correctly."""
    export(source, tmp_path / target)
    export(source, tmp_path / target, exist_ok=True)
    with pytest.raises(ExistsError):
        export(source, tmp_path / target, exist_ok=False)
    assert (tmp_path / written).read_text("utf-8") == "\n\n".join(FUNCS.values())


@pytest.mark.parametrize(
    ("source", "include", "exclude"),
    [
        (NBPATH, None, None),
        (NBPATH, ["another_func", "print_this"], None),
        (NBPATH, None, ["another_func", "print_this"]),
    ],
)
def test_extract(
    source: Path,
    include: list[str] | None,
    exclude: list[str] | None,
) -> None:
    """Extract functions from a source file."""
    _keep = include or FUNCS.keys()
    _remove = exclude or []
    assert set(extract(src=source, include=include, exclude=exclude)) == {
        Definition(name=fname, src=fsrc, path=NBPATH)
        for fname, fsrc in FUNCS.items()
        if fname in _keep and fname not in _remove
    }


def test_errors(tmp_path: Path) -> None:
    """Raise errors."""
    with pytest.raises(FileSuffixError):
        export(NBPATH, tmp_path / "target.txt")
    with pytest.raises(PathNameError):
        export(NBPATH, tmp_path / "target")
    with pytest.raises(NotFoundError):
        extract(src="NBPATH")

    with pytest.raises(InputsError):
        extract(src=NBPATH, include=[], exclude=[])
    with pytest.raises(InputsError):
        extract(src=NBPATH.parents[1])

    export(NBPATH, tmp_path / "target.py")
    with pytest.raises(ExistsError):
        export(NBPATH, tmp_path / "target.py", exist_ok=False)


def test__combine_funcs_update() -> None:
    """Combine functions from source and destinations."""
    foo = Path("foo.py")
    bar = Path("bar.py")
    src = [
        Definition(
            name="another_func",
            path=bar,
            src="def another_func() -> None:\n  ...",
        ),
        Definition(
            name="print_this",
            path=foo,
            src="def print_this(x: str) -> None:\n  print(x)",
        ),
    ]
    dest = [
        Definition(
            name="print_that",
            path=foo,
            src="def print_this(x: str) -> None:\n  print('out:', x)",
        ),
        Definition(
            name="another_func",
            path=bar,
            src="def another_func() -> None:\n  return",
        ),
    ]
    assert set(_combine_funcs(src=src, dest=dest, update=None)) == {
        Definition(
            name="print_that",
            path=foo,
            src="def print_this(x: str) -> None:\n  print('out:', x)",
        ),
        Definition(
            name="print_this",
            path=foo,
            src="def print_this(x: str) -> None:\n  print(x)",
        ),
        Definition(
            name="another_func",
            path=bar,
            src="def another_func() -> None:\n  return",
        ),
    }
    assert set(_combine_funcs(src=src, dest=dest, update=True)) == {
        Definition(
            name="another_func",
            path=bar,
            src="def another_func() -> None:\n  return",
        ),
        Definition(
            name="print_that",
            path=foo,
            src="def print_this(x: str) -> None:\n  print('out:', x)",
        ),
    }



"""CLI utilities."""
from __future__ import annotations

import argparse

from nbdefs2py.io import export

parser = argparse.ArgumentParser(
    prog="python -m nbfuncs",
    description="Extract definitions from notebooks.",
)
parser.add_argument("source", metavar="SRC", type=str, help="source file/path")
parser.add_argument("destination", metavar="DST", type=str, help="target file/path")
parser.add_argument(
    "-i",
    "--ignore",
    type=str,
    default=None,
    help="glob expression of files to ignore",
)

parser.add_argument(
    "--update",
    action="store_true",
    help="update only existing functions",
)
parser.add_argument(
    "--no-update",
    dest="update",
    action="store_false",
    help="overwrite destination file",
)
parser.add_argument(
    "--include",
    type=str,
    nargs="+",
    default=None,
    help="names of functions to include",
)
parser.add_argument(
    "--exclude",
    type=str,
    nargs="+",
    default=None,
    help="names of functions to ignore",
)


def main(args: list[str] | None = None) -> None:
    """Execute from CLI arguments."""
    parser.set_defaults(update=None)
    parsed_args = parser.parse_args(args)
    export(**vars(parsed_args))


if __name__ == "__main__":
    main()
